#!/usr/bin/env python
"""Helper script to run reports in the command line"""

import ast
import logging

from tlbx import Script, Arg, raiseif, st

from zillion.core import info, set_log_level, nlp_installed
from zillion.datasource import DataSource
from zillion.warehouse import Warehouse


@Script(
    Arg("config", help="Path to warehouse config file"),
    Arg(
        "-ds",
        "--ds-config",
        action="store_true",
        default=False,
        help="Interpret the config as a DataSource config and create a Warehouse from the DataSource",
    ),
    Arg("-m", "--metrics", nargs="+", help="Metrics to include in report"),
    Arg("-d", "--dimensions", nargs="+", help="Dimensions to include in report"),
    Arg("-c", "--criteria", help="String to be eval'd as criteria param"),
    Arg("-o", "--order-by", help="String to be eval'd as order by param"),
    Arg("-r", "--rollup", help="Rollup setting"),
    Arg("-rf", "--row-filters", help="String to be eval'd as row filter param"),
    Arg("-l", "--limit", type=int, help="Limit setting"),
    Arg(
        "-t",
        "--text",
        type=str,
        help="Execute a report from natural language text. Requires the NLP extension.",
    ),
    Arg("-ll", "--log-level", type=int, default=logging.INFO, help="Set log level"),
)
def main(
    config=None,
    ds_config=False,
    metrics=None,
    dimensions=None,
    criteria=None,
    order_by=None,
    row_filters=None,
    rollup=None,
    limit=None,
    text=None,
    log_level=None,
):
    if log_level:
        set_log_level(log_level)

    raiseif(
        text and not nlp_installed,
        "The NLP extension is not installed. Please install it with `pip install zillion[nlp]` to use the --text option.",
    )

    if ds_config:
        ds = DataSource("bootstrap", config=config)
        wh = Warehouse(datasources=[ds])
    else:
        wh = Warehouse(config=config)

    if text:
        info(f"Executing report from text: {text}")
        result = wh.execute_text(text)
    else:
        if criteria:
            criteria = ast.literal_eval(criteria)
        if order_by:
            order_by = ast.literal_eval(order_by)
        if row_filters:
            row_filters = ast.literal_eval(row_filters)

        params = dict(
            metrics=metrics,
            dimensions=dimensions,
            criteria=criteria,
            order_by=order_by,
            row_filters=row_filters,
            rollup=rollup,
            limit=limit,
        )
        info("Executing report with params:")
        info(params)
        result = wh.execute(**params)

    info(result.df_display)


if __name__ == "__main__":
    main()



# pylint: disable=unused-import,missing-class-docstring
from collections.abc import MutableMapping
import logging
import os
import requests
import sys
import time

nlp_installed = False

try:
    import langchain
    import qdrant_client

    nlp_installed = True
except ImportError:
    pass

from tlbx import (
    st,
    dbg as _dbg,
    dbgsql as _dbgsql,
    info as _info,
    warn as _warn,
    error as _error,
    pf,
    pp,
    format_msg,
    sqlformat,
    rmfile,
    get_caller,
    open_filepath_or_buffer,
    get_string_format_args,
    json,
    chunks,
    iter_or,
    powerset,
    is_int,
    orderedsetify,
    initializer,
    PrintMixin,
    MappingMixin,
    ClassValueContainsMeta,
    get_class_var_values,
    get_class_vars,
    import_object,
)
import yaml

# Last unicode char - this helps get the rollup rows to sort last, but may
# need to be replaced for presentation.
ROLLUP_INDEX_LABEL = chr(1114111)
# HACK: pandas can't group MultiIndex NaN values, so we replace them with a
# value we *hope* to never see in the index as a workaround.
NAN_DIMENSION_VALUE_LABEL = chr(1114110)
# This is more friendly for front-end viewing, but has a better chance of
# conflicting with actual report data.
ROLLUP_INDEX_DISPLAY_LABEL = "::"

ADHOC_DS_URL = "adhoc"  # A placeholder to denote its an adhoc datasource
RESERVED_FIELD_NAMES = set(["row_hash"])
DEFAULT_REPLACE_AFTER = "1 days"
CRITERIA_OPERATIONS = set(
    [
        ">",
        ">=",
        "<",
        "<=",
        "=",
        "!=",
        "in",
        "not in",
        "between",
        "not between",
        "like",
        "not like",
    ]
)
ROW_FILTER_OPERATIONS = set([">", ">=", "<", "<=", "=", "!="])


class ZillionException(Exception):
    pass


class InvalidTechnicalException(ZillionException):
    pass


class WarehouseException(ZillionException):
    pass


class InvalidWarehouseIdException(ZillionException):
    pass


class ReportException(ZillionException):
    pass


class InvalidReportIdException(ZillionException):
    pass


class UnsupportedGrainException(ZillionException):
    pass


class UnsupportedKillException(ZillionException):
    pass


class FailedKillException(ZillionException):
    pass


class DataSourceQueryTimeoutException(ZillionException):
    pass


class ExecutionKilledException(ZillionException):
    pass


class ExecutionLockException(ZillionException):
    pass


class InvalidFieldException(ZillionException):
    pass


class InvalidDimensionValueException(ZillionException):
    pass


class DisallowedSQLException(ZillionException):
    pass


class MaxFormulaDepthException(ZillionException):
    pass


class FieldTypes(metaclass=ClassValueContainsMeta):
    """Allowed field types"""

    DIMENSION = "dimension"
    METRIC = "metric"


class TableTypes(metaclass=ClassValueContainsMeta):
    """Allowed table types"""

    DIMENSION = "dimension"
    METRIC = "metric"


class AggregationTypes(metaclass=ClassValueContainsMeta):
    """Allowed aggregation types. These aggregations are limited by what can be
    done in most SQL databases"""

    MEAN = "mean"
    SUM = "sum"
    MIN = "min"
    MAX = "max"
    COUNT = "count"
    COUNT_DISTINCT = "count_distinct"


class TechnicalTypes(metaclass=ClassValueContainsMeta):
    """Allowed technical types"""

    MEAN = "mean"
    SUM = "sum"
    MEDIAN = "median"
    STD = "std"
    VAR = "var"
    MIN = "min"
    MAX = "max"
    BOLL = "boll"
    DIFF = "diff"
    PCT_CHANGE = "pct_change"
    CUMSUM = "cumsum"
    CUMMIN = "cummin"
    CUMMAX = "cummax"
    RANK = "rank"
    PCT_RANK = "pct_rank"


class TechnicalModes(metaclass=ClassValueContainsMeta):
    """Allowed Technical modes

    **Attributes:**

    * **GROUP** - (*str*) Apply the technical to the last grouping of the data
    for a multi-dimensional report
    * **ALL** - (*str*) Apply the technical across all result data

    """

    GROUP = "group"
    ALL = "all"


class RollupTypes(metaclass=ClassValueContainsMeta):
    """Allowed Rollup Types"""

    TOTALS = "totals"
    ALL = "all"


class OrderByTypes(metaclass=ClassValueContainsMeta):
    """Allowed Order By Types"""

    ASC = "asc"
    DESC = "desc"


class DataSourceQueryModes(metaclass=ClassValueContainsMeta):
    """Allowed datasource query modes"""

    SEQUENTIAL = "sequential"
    MULTITHREAD = "multithread"


class ExecutionState:
    """Allowed report/query execution states"""

    READY = "ready"
    QUERYING = "querying"
    KILLED = "killed"


class IfExistsModes(metaclass=ClassValueContainsMeta):
    """Allowed modes when creating tables from data. This is based off of pandas
    `if_exists` param in the `DataFrame.to_sql` method, with the addition of an
    "ignore" option. The "append" option is also removed for now since there
    isn't a safe/generic way to guarantee a proper primary key has been set on
    the table."""

    FAIL = "fail"
    REPLACE = "replace"
    # APPEND = "append"
    IGNORE = "ignore"


class IfFileExistsModes(IfExistsModes):
    """An extension of the modes above specific to downloaded files. This
    allows the config to specify that a downloaded file should be replaced
    after a certain amount of time. See code that uses this for implementation
    details.
    """

    REPLACE_AFTER = "replace_after"


# ---- TODO: move below to utils file


def raiseif(cond, msg="", exc=ZillionException):
    """Convenience assert-like utility"""
    if cond:
        raise exc(msg)


def raiseifnot(cond, msg="", exc=ZillionException):
    """Convenience assert-like utility"""
    if not cond:
        raise exc(msg)


def igetattr(obj, attr, *args):
    """Case-insensitive getattr"""
    for a in dir(obj):
        if a.lower() == attr.lower():
            return getattr(obj, a)
    if args:
        return args[0]
    raise AttributeError("type object '%s' has no attribute '%s'" % (type(obj), attr))


def read_filepath_or_buffer(f, open_flags="r", compression=None):
    """Open and read files or buffers, local or remote"""
    f, handles, close = open_filepath_or_buffer(
        f, open_flags=open_flags, compression=compression
    )
    try:
        data = f.read()
    finally:
        if close:
            try:
                f.close()
            except ValueError:
                pass
    return data


def download_file(url, outfile=None):
    """Utility to download a datafile"""
    if not outfile:
        outfile = url.split("/")[-1]
    info("Downloading %s to %s" % (url, outfile))
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(outfile, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return outfile


def get_modified_time(fname):
    """Utility to get the modified time of a file"""
    return os.stat(fname).st_mtime


def get_time_since_modified(fname):
    """Utility to get the time since a file was last modified"""
    return time.time() - get_modified_time(fname)


def load_yaml(fname):
    """Wrapper to safe_load that also expands environment vars"""
    with open(fname) as f:
        val = yaml.safe_load(os.path.expandvars(f.read()))
    return val


def load_json_or_yaml_from_str(string, f=None, schema=None):
    """Load the file as json or yaml depending on the extension (if f is a string)
    or by trying both (if f is not a string). If you know ahead of time that your
    data is one or the other, you should use yaml or json load directly.

    **Parameters:**

    * **string** - (*str*) The raw json or yaml string
    * **f** - (*str or buffer*) A file path or buffer where contents were read from
    * **schema** - (*optional*) Validate against this schema

    **Returns:**

    (*dict*) - A dict structure loaded from the json/yaml

    """

    load_result = None
    if f and isinstance(f, str):  # f is assumed to be a filename
        f = f.lower()
        if f.endswith("yaml") or f.endswith("yml"):
            load_result = yaml.safe_load(string)
        elif f.endswith("json"):
            load_result = json.loads(string)

    if load_result is None:
        # f is not a filename, we try json and then fall back to yaml
        try:
            load_result = json.loads(string)
        except Exception as ej:
            try:
                load_result = yaml.safe_load(string)
            except Exception as ey:
                raise Exception("Could not load string as json or yaml")

    if schema:
        return schema.load(load_result)
    return load_result


def dictmerge(x, y, path=None, overwrite=False, extend=False):
    """Adapted version of tlbx's dictmerge that supports extending lists"""
    if path is None:
        path = []
    for key in y:
        if key in x:
            if isinstance(x[key], (dict, MutableMapping)) and isinstance(
                y[key], (dict, MutableMapping)
            ):
                dictmerge(
                    x[key],
                    y[key],
                    path + [str(key)],
                    overwrite=overwrite,
                    extend=extend,
                )
            elif x[key] == y[key]:
                pass  # same leaf value
            else:
                if not overwrite:
                    raise Exception("Conflict at %s" % ".".join(path + [str(key)]))
                if isinstance(x[key], list) and isinstance(y[key], list) and extend:
                    x[key].extend(y[key])
                else:
                    x[key] = y[key]
        else:
            x[key] = y[key]
    return x


def load_zillion_config():
    """If the ZILLION_CONFIG environment variable is defined, read the YAML
    config from this file. Environment variable substitution is supported
    in the yaml file. Otherwise return a default config. Environment variables
    prefixed with "ZILLION_"  will also be read in (with the prefix stripped)
    and take precedence.

    **Returns:**

    (*dict*) - The zillion config dict.

    """
    zillion_config_fname = os.environ.get("ZILLION_CONFIG", None)
    if zillion_config_fname:
        # Load with support for filling in env var values
        config = load_yaml(zillion_config_fname)
        for k, v in config.copy().items():
            # Hack: some older config items had ZILLION prefixed. As a workaround
            # we now always remove that prefix and map to the key without the prefix.
            if k.startswith("ZILLION_"):
                config[k.replace("ZILLION_", "")] = v
                del config[k]
    else:
        print("No ZILLION_CONFIG specified, using default settings")
        config = dict(
            DEBUG=False,
            LOG_LEVEL="WARNING",
            DB_URL="sqlite:////tmp/zillion.db",
            ADHOC_DATASOURCE_DIRECTORY="/tmp",
            LOAD_TABLE_CHUNK_SIZE=5000,
            OPENAI_API_KEY=None,
            OPENAI_MODEL="gpt-3.5-turbo",
            QDRANT_HOST=None,
            DATASOURCE_QUERY_MODE=DataSourceQueryModes.SEQUENTIAL,
            DATASOURCE_QUERY_TIMEOUT=None,
            DATASOURCE_CONTEXTS={},
        )

    for k, v in os.environ.items():
        if k.startswith("ZILLION_"):
            config[k.replace("ZILLION_", "")] = v

    if isinstance(config.get("DATASOURCE_CONTEXTS", None), str):
        config["DATASOURCE_CONTEXTS"] = load_json_or_yaml_from_str(
            config["DATASOURCE_CONTEXTS"]
        )

    return config


zillion_config = load_zillion_config()


def get_zillion_config_log_level():
    return getattr(logging, zillion_config.get("LOG_LEVEL", "WARNING").upper())


default_logger = logging.getLogger("zillion")


def set_log_level_from_config(cfg):
    global default_logger
    if str(cfg.get("DEBUG", "false")).lower() in ("true", "1"):
        default_logger.setLevel(logging.DEBUG)
        # Make sure logs can show up in testing
        handler = logging.StreamHandler(sys.stdout)
        default_logger.handlers = []
        default_logger.propagate = False
        default_logger.addHandler(handler)
        print("---- Zillion debug logging enabled ----")
    else:
        default_logger.setLevel(get_zillion_config_log_level())


set_log_level_from_config(zillion_config)


def set_log_level(level):
    """Set the log level for the zillion logger"""
    global default_logger
    default_logger.setLevel(level)


def dbg(msg, **kwargs):
    """Call tlbx dbg with zillion logger"""
    kwargs["logger"] = kwargs.get("logger", default_logger)
    kwargs["label"] = kwargs.get("label", get_caller())
    _dbg(msg, **kwargs)


def dbgsql(msg, **kwargs):
    """Call tlbx dbgsql with zillion logger"""
    kwargs["logger"] = kwargs.get("logger", default_logger)
    kwargs["label"] = kwargs.get("label", get_caller())
    _dbgsql(msg, **kwargs)


def info(msg, **kwargs):
    """Call tlbx info with zillion logger"""
    kwargs["logger"] = kwargs.get("logger", default_logger)
    kwargs["label"] = kwargs.get("label", get_caller())
    _info(msg, **kwargs)


def warn(msg, **kwargs):
    """Call tlbx warn with zillion logger"""
    kwargs["logger"] = kwargs.get("logger", default_logger)
    kwargs["label"] = kwargs.get("label", get_caller())
    _warn(msg, **kwargs)


def error(msg, **kwargs):
    """Call tlbx error with zillion logger"""
    kwargs["logger"] = kwargs.get("logger", default_logger)
    kwargs["label"] = kwargs.get("label", get_caller())
    _error(msg, **kwargs)



import sqlalchemy as sa
from sqlalchemy import func
from sqlalchemy.dialects.postgresql import DATE, TIMESTAMP
from sqlalchemy.sql.elements import Grouping
from tlbx import st

from zillion.dialects.conversions import DialectDateConversions

# Have to add all the sa.literals to avoid compile errors while printing
# queries: https://github.com/sqlalchemy/sqlalchemy/issues/5859


class DuckDBDialectDateConversions(DialectDateConversions):
    @classmethod
    def date_year_start(cls, x):
        return func.CAST(sa.literal(str(x) + "-01-01"), DATE)

    @classmethod
    def date_year_plus_year(cls, x):
        return Grouping(
            func.CAST(sa.literal(str(x) + "-01-01"), DATE) + func.to_years(1)
        )

    @classmethod
    def datetime_year_end(cls, x):
        return Grouping(
            func.CAST(sa.literal(str(x) + "-01-01"), TIMESTAMP)
            + func.to_years(1)
            - func.to_seconds(1)
        )

    @classmethod
    def date_month_start(cls, x):
        return func.strftime(func.CAST(sa.literal(str(x) + "-01"), DATE), "%Y-%m")

    @classmethod
    def date_month_plus_month(cls, x):
        return func.strftime(
            func.CAST(sa.literal(str(x) + "-01"), DATE) + func.to_months(1), "%Y-%m"
        )

    @classmethod
    def datetime_month_end(cls, x):
        return Grouping(
            func.CAST(sa.literal(str(x) + "-01"), TIMESTAMP)
            + func.to_months(1)
            - func.to_seconds(1)
        )

    @classmethod
    def date_plus_day(cls, x):
        return Grouping(func.CAST(sa.literal(str(x)), DATE) + func.to_days(1))

    @classmethod
    def datetime_day_end(cls, x):
        return Grouping(
            func.CAST(sa.literal(str(x)), TIMESTAMP)
            + func.to_days(1)
            - func.to_seconds(1)
        )

    @classmethod
    def datetime_hour_plus_hour(cls, x):
        return Grouping(func.CAST(sa.literal(str(x)), TIMESTAMP) + func.to_hours(1))

    @classmethod
    def datetime_hour_end(cls, x):
        return Grouping(
            func.CAST(sa.literal(str(x)), TIMESTAMP)
            + func.to_hours(1)
            - func.to_seconds(1)
        )

    @classmethod
    def datetime_minute_plus_minute(cls, x):
        return Grouping(func.CAST(sa.literal(str(x)), TIMESTAMP) + func.to_minutes(1))

    @classmethod
    def datetime_minute_end(cls, x):
        return Grouping(
            func.CAST(sa.literal(str(x)), TIMESTAMP)
            + func.to_minutes(1)
            - func.to_seconds(1)
        )


DUCKDB_DIALECT_CONVERSIONS = {
    "year": {
        "ds_formula": "EXTRACT(YEAR FROM {})",
        "ds_criteria_conversions": DuckDBDialectDateConversions.get_year_criteria_conversions(),
    },
    "quarter": "strftime({}, '%Y-Q')  || date_part('quarter', {})",
    "quarter_of_year": "EXTRACT(QUARTER FROM {})",
    "month": {
        "ds_formula": "strftime({}, '%Y-%m')",
        "ds_criteria_conversions": DuckDBDialectDateConversions.get_month_criteria_conversions(),
    },
    "month_name": "strftime({}, '%B')",
    "month_of_year": "EXTRACT(MONTH FROM {})",
    "week_of_month": "EXTRACT(WEEK FROM {}) - EXTRACT(WEEK FROM CAST(DATE_TRUNC('month', {}) as date)) + 1",
    "week_of_year": "EXTRACT(WEEK FROM {})",  # HACK: attempt to get tests compatible with mysql and sqlite
    "period_of_month_7d": "FLOOR((EXTRACT(DAY FROM {}) - 1) / 7) + 1",
    "date": {
        "ds_formula": "strftime({}, '%Y-%m-%d')",
        "ds_criteria_conversions": DuckDBDialectDateConversions.get_date_criteria_conversions(),
    },
    "day_name": "strftime({}, '%A')",
    "day_of_week": "EXTRACT(ISODOW FROM {})",  # Monday = 1
    "is_weekday": (
        "CASE EXTRACT(ISODOW FROM {}) "
        "WHEN 1 THEN 1 "
        "WHEN 2 THEN 1 "
        "WHEN 3 THEN 1 "
        "WHEN 4 THEN 1 "
        "WHEN 5 THEN 1 "
        "WHEN 6 THEN 0 "
        "WHEN 7 THEN 0 "
        "ELSE NULL "
        "END"
    ),
    "day_of_month": "EXTRACT(DAY FROM {})",
    "day_of_year": "EXTRACT(DOY FROM {})",
    "hour": {
        "ds_formula": "strftime({}, '%Y-%m-%d %H:00:00')",
        "ds_criteria_conversions": DuckDBDialectDateConversions.get_hour_criteria_conversions(),
    },
    "hour_of_day": "EXTRACT(HOUR FROM {})",
    "minute": {
        "ds_formula": "strftime({}, '%Y-%m-%d %H:%M:00')",
        "ds_criteria_conversions": DuckDBDialectDateConversions.get_minute_criteria_conversions(),
    },
    "minute_of_hour": "EXTRACT(MINUTE FROM {})",
    "datetime": {
        "ds_formula": "strftime({}, '%Y-%m-%d %H:%M:%S')",
        "ds_criteria_conversions": DuckDBDialectDateConversions.get_datetime_criteria_conversions(),
    },
    "unixtime": "EXTRACT(epoch from {})",
}



from collections import defaultdict, OrderedDict
import time

import sqlalchemy as sa

from zillion.configs import load_warehouse_config, is_active
from zillion.core import *
from zillion.datasource import DataSource
from zillion.field import get_table_dimensions, get_table_fields, FieldManagerMixin
from zillion.model import zillion_engine, Warehouses
from zillion.nlp import init_warehouse_embeddings
from zillion.report import Report
from zillion.sql_utils import is_numeric_type, column_fullname


class Warehouse(FieldManagerMixin):
    """A reporting warehouse that contains various datasources to run queries
    against and combine data in report results. The warehouse may contain global
    definitions for metrics and dimensions, and will also perform integrity
    checks of any added datasources.

    Note that the id, name, and meta attributes will only be populated when
    the Warehouse is persisted or loaded from a database.

    **Parameters:**

    * **config** - (*dict, str, or buffer, optional*) A dict adhering to the
    WarehouseConfigSchema or a file location to load the config from
    * **datasources** - (*list, optional*) A list of DataSources that will make
    up the warehouse
    * **ds_priority** - (*list, optional*) An ordered list of datasource names
    establishing querying priority. This comes into play when part of a report
    may be satisfied by multiple datasources. Datasources earlier in this list
    will be higher priority.
    * **nlp** - (*bool, optional*) If true, allow NLP analysis when creating
    fields

    """

    def __init__(self, config=None, datasources=None, ds_priority=None, nlp=False):
        # Note: id/name get updated on save/load
        self.id = None
        self.name = None
        self.meta = None
        self.nlp = nlp
        self._datasources = OrderedDict()
        self._metrics = {}
        self._dimensions = {}
        self._supported_dimension_cache = {}

        for ds in datasources or []:
            self.add_datasource(ds, skip_integrity_checks=True)

        if config:
            config = load_warehouse_config(config)
            self.apply_config(config, skip_integrity_checks=True, nlp=nlp)

        raiseifnot(self._datasources, "No datasources provided or found in config")

        self._add_default_display_names()
        self.run_integrity_checks()

        self.ds_priority = ds_priority or list(self._datasources.keys())

        raiseifnot(
            isinstance(self.ds_priority, list),
            (
                "Invalid format for ds_priority, must be list of datasource names: %s"
                % self.ds_priority
            ),
        )
        raiseifnot(
            len(self.ds_priority) == len(self._datasources),
            "Length mismatch between ds_priority and datasources",
        )
        for ds_name in self.ds_priority:
            raiseifnot(
                ds_name in self._datasources,
                "Datasource %s is in ds_priority but not in datasource map" % ds_name,
            )

    def __repr__(self):
        return "<%s(name=%s)> Datasources: %s" % (
            self.__class__.__name__,
            self.name,
            self.datasource_names,
        )

    @property
    def datasources(self):
        """The datasource objects in this warehouse"""
        return self._datasources.values()

    @property
    def datasource_names(self):
        """The names of datasources in this warehouse"""
        return self._datasources.keys()

    def print_info(self):
        """Print the warehouse structure"""
        print("---- Warehouse")
        print("metrics:")
        self.print_metrics(indent=2)
        print("dimensions:")
        self.print_dimensions(indent=2)

        for ds in self.datasources:
            print()
            ds.print_info()

    def get_datasource(self, name, adhoc_datasources=None):
        """Get the datasource object corresponding to this datasource name

        **Parameters:**

        * **name** - (*str*) The name of the datasource
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers
        specific to this request

        **Returns:**

        (*DataSource*) - The matching datasource object

        """
        if name in self._datasources:
            return self._datasources[name]

        for adhoc_datasource in adhoc_datasources or []:
            if adhoc_datasource.name == name:
                return adhoc_datasource

        raise ZillionException('Could not find datasource with name "%s"' % name)

    def get_child_field_managers(self):
        """Get a list of all datasources in this warehouse"""
        return list(self.datasources)

    def add_datasource(self, ds, skip_integrity_checks=False):
        """Add a datasource to this warehouse

        **Parameters:**

        * **ds** - (*DataSource*) The datasource object to add
        * **skip_integrity_checks** - (*bool, optional*) If True, skip warehouse
        integrity checks

        """
        dbg("Adding datasource %s" % ds.name)
        self._datasources[ds.name] = ds
        self._clear_supported_dimension_cache()
        if not skip_integrity_checks:
            self.run_integrity_checks()

    def remove_datasource(self, ds, skip_integrity_checks=False):
        """Remove a datasource from this config

        **Parameters:**

        * **ds** - (*DataSource*) The datasource object to remove
        * **skip_integrity_checks** - (*bool, optional*) If True, skip warehouse
        integrity checks

        """
        dbg("Removing datasource %s" % ds.name)
        del self._datasources[ds.name]
        self._clear_supported_dimension_cache()
        if not skip_integrity_checks:
            self.run_integrity_checks()

    def apply_config(self, config, skip_integrity_checks=False, nlp=False):
        """Apply a warehouse config

        **Parameters:**

        * **config** - (*dict*) A dict adhering to the WarehouseConfigSchema
        * **skip_integrity_checks** - (*bool, optional*) If True, skip warehouse
        integrity checks
         * **nlp** - (*bool, optional*) If true, allow NLP analysis when creating
        fields

        """
        self._create_or_update_datasources(
            config.get("datasources", {}),
            skip_integrity_checks=skip_integrity_checks,
            nlp=nlp,
        )

        if config.get("meta", None):
            meta = self.meta or {}
            # The config meta takes precedence over any existing meta settings
            self.meta = dictmerge(meta, config["meta"], overwrite=True)

        # NOTE: this goes second in case any formula fields reference fields
        # defined or created in the datasources. It may make more sense to
        # only defer population of formula fields.
        self._populate_global_fields(config, force=True)
        self._clear_supported_dimension_cache()

    def run_integrity_checks(self, adhoc_datasources=None):
        """Run a series of integrity checks on the warehouse and its
        datasources. This will raise a WarehouseException with all failed
        checks.

        **Parameters:**

        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers to
        include for this request

        """
        errors = []
        if adhoc_datasources:
            for ds in adhoc_datasources:
                if ds.name in self.datasource_names:
                    errors.append(
                        "Adhoc DataSource '%s' name conflicts with existing DataSource"
                        % ds.name
                    )

        errors.extend(
            self._check_reserved_field_names(adhoc_datasources=adhoc_datasources)
        )
        errors.extend(
            self._check_conflicting_fields(adhoc_datasources=adhoc_datasources)
        )
        errors.extend(self._check_fields_have_type(adhoc_datasources=adhoc_datasources))
        errors.extend(
            self._check_primary_key_dimensions(adhoc_datasources=adhoc_datasources)
        )
        errors.extend(
            self._check_weighting_metrics(adhoc_datasources=adhoc_datasources)
        )
        errors.extend(self._check_required_grain(adhoc_datasources=adhoc_datasources))
        errors.extend(
            self._check_incomplete_dimensions(adhoc_datasources=adhoc_datasources)
        )
        errors.extend(
            self._check_valid_table_parents(adhoc_datasources=adhoc_datasources)
        )

        if errors:
            raise WarehouseException("Integrity check(s) failed.\n%s" % pf(errors))

    def load_report(self, spec_id, adhoc_datasources=None):
        """Load a report from a spec ID

        **Parameters:**

        * **spec_id** - (*int*) The ID of a report spec
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers
        specific to this request

        **Returns:**

        (*Report*) - A report built from this report spec

        """
        raiseifnot(
            self.id,
            "The Warehouse must be saved before ReportSpecs can be loaded for the Warehouse",
        )
        return Report.load(self, spec_id, adhoc_datasources=adhoc_datasources)

    def delete_report(self, spec_id):
        """Delete a report by spec ID

        **Parameters:**

        * **spec_id** - (*int*) The ID of a report spec to delete

        """
        raiseifnot(
            self.id,
            "The Warehouse must be saved before ReportSpecs can be deleted for the Warehouse",
        )
        Report.delete(self, spec_id)

    def save_report(self, meta=None, **kwargs):
        """Init a Report and save it as a ReportSpec. Note that the Warehouse
        must be saved before any ReportSpecs can be saved for the Warehouse.

        **Parameters:**

        * **meta** - (*object, optional*) A metadata object to be
        serialized as JSON and stored with the report
        * ****kwargs** - Passed through to Report

        **Returns:**

        (*Report*) - The built report with the spec ID populated

        """
        raiseifnot(
            self.id,
            "The Warehouse must be saved before ReportSpecs can be saved for the Warehouse",
        )
        report = Report(self, **kwargs)
        report.save(meta=meta)
        return report

    def save(self, name, config_url, meta=None):
        """Save the warehouse config and return the ID

        **Parameters:**

        * **name** - (*str*) A name to give the Warehouse
        * **config_url** - (*str*) A URL pointing to a config file that can
        be used to recreate the warehouse
        * **meta** - (*object, optional*) A metadata object to be
        serialized as JSON and stored with the warehouse. This will be merged
        into existing meta settings if present and take precedence.

        **Returns:**

        (*int*) - The ID of the saved Warehouse

        """
        raiseifnot(name, "A unique name must be specified to save a Warehouse")
        raiseifnot(
            # TODO: better check for valid URL
            config_url and isinstance(config_url, str),
            "A config URL must be specified to save a Warehouse",
        )

        params = dict(ds_priority=self.ds_priority, config=config_url, nlp=self.nlp)

        self.meta = dictmerge(self.meta or {}, meta or {}, overwrite=True)

        conn = zillion_engine.connect()
        try:
            result = conn.execute(
                Warehouses.insert(),
                name=name,
                params=json.dumps(params),
                meta=json.dumps(self.meta),
            )
            wh_id = result.inserted_primary_key[0]
            raiseifnot(wh_id, "No warehouse ID found")
        finally:
            conn.close()

        self.id = wh_id
        self.name = name
        return wh_id

    def execute(
        self,
        metrics=None,
        dimensions=None,
        criteria=None,
        row_filters=None,
        rollup=None,
        pivot=None,
        order_by=None,
        limit=None,
        limit_first=False,
        adhoc_datasources=None,
        allow_partial=False,
    ):
        """Build and execute a Report

        **Returns:**

        (*ReportResult*) - The result of the report

        """
        start = time.time()

        report = Report(
            self,
            metrics=metrics,
            dimensions=dimensions,
            criteria=criteria,
            row_filters=row_filters,
            rollup=rollup,
            pivot=pivot,
            order_by=order_by,
            limit=limit,
            limit_first=limit_first,
            adhoc_datasources=adhoc_datasources,
            allow_partial=allow_partial,
        )
        result = report.execute()

        dbg("warehouse report took %.3fs" % (time.time() - start))
        return result

    def execute_id(self, spec_id, adhoc_datasources=None):
        """Build and execute a report from a spec ID

        **Parameters:**

        * **spec_id** - (*int*) The ID of a report spec
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers
        specific to this request

        **Returns:**

        (*ReportResult*) - The result of the report

        """
        start = time.time()
        report = self.load_report(spec_id, adhoc_datasources=adhoc_datasources)
        result = report.execute()
        dbg("warehouse report took %.3fs" % (time.time() - start))
        return result

    def execute_text(self, text, adhoc_datasources=None, allow_partial=False):
        """Build and execute a report from a natural language query. Requires
        the nlp extension to be installed.

        **Parameters:**

        * **text** - (*str*) A natural language query
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers
        specific to this request
        * **allow_partial** - (*bool, optional*) If True, allow partial

        **Returns:**

        (*ReportResult*) - The result of the report

        """
        if not nlp_installed:
            raise WarehouseException("nlp extension is not installed")

        if not self._get_embeddings_collection_name():
            info(f"Initializing Warehouse embeddings")
            self.init_embeddings()

        start = time.time()
        report = Report.from_text(
            self, text, adhoc_datasources=adhoc_datasources, allow_partial=allow_partial
        )
        result = report.execute()
        dbg("warehouse report took %.3fs" % (time.time() - start))
        return result

    def get_metric_table_set(
        self, metric, grain, dimension_grain, adhoc_datasources=None
    ):
        """Get a TableSet that can satisfy a metric at a given grain

        **Parameters:**

        * **metric** - (*str*) A metric name
        * **grain** - (*list*) A list of dimension names representing the full
        grain required including dimension and criteria grain
        * **dimension_grain** - (*list of str*) A list of dimension names
        representing the requested dimensions for report grouping
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers for
        this request

        **Returns:**

        (*TableSet*) - A TableSet that can satisfy this request

        """
        dbg("metric:%s grain:%s" % (metric, grain))
        ds_metric_tables = self._get_ds_tables_with_metric(
            metric, adhoc_datasources=adhoc_datasources
        )
        ds_table_sets = self._get_ds_table_sets(
            ds_metric_tables,
            metric,
            grain,
            dimension_grain,
            adhoc_datasources=adhoc_datasources,
        )
        if not ds_table_sets:
            msg = self._generate_unsupported_grain_msg(
                grain, metric, adhoc_datasources=adhoc_datasources
            )
            raise UnsupportedGrainException(msg)
        table_set = self._choose_best_table_set(ds_table_sets)
        return table_set

    def get_dimension_table_set(self, grain, dimension_grain, adhoc_datasources=None):
        """Get a TableSet that can satisfy dimension table joins across this
        grain

        **Parameters:**

        * **grain** - (*list*) A list of dimension names representing the full
        grain required including dimension and criteria grain
        * **dimension_grain** - (*list of str*) A list of dimension names
        representing the requested dimensions for report grouping
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers for
        this request

        **Returns:**

        (*TableSet*) - A TableSet that can satisfy this request

        """
        dbg("grain:%s" % grain)

        table_set = None
        for dim_name in grain:
            ds_dim_tables = self._get_ds_dim_tables_with_dim(
                dim_name, adhoc_datasources=adhoc_datasources
            )
            ds_table_sets = self._get_ds_table_sets(
                ds_dim_tables,
                dim_name,
                grain,
                dimension_grain,
                adhoc_datasources=adhoc_datasources,
            )
            if not ds_table_sets:
                continue
            table_set = self._choose_best_table_set(ds_table_sets)
            break

        if not table_set:
            raise UnsupportedGrainException(
                "No dimension table set found to meet grain: %s" % grain
            )
        return table_set

    def init_embeddings(self, force_recreate=False):
        """Initialize the warehouse embeddings collection. This is necessary
        to use natural language query to report features.

        **Parameters:**

        * **force_recreate** - (*bool, optional*) If True, force the embeddings
        collection to be recreated from scratch.

        """
        collection_name = init_warehouse_embeddings(self, force_recreate=force_recreate)
        self._set_embeddings_collection_name(collection_name)

    def _get_embeddings_collection_name(self):
        """Get the name of the embeddings collection for this warehouse"""
        return (self.meta or {}).get("nlp", {}).get("collection_name", None)

    def _set_embeddings_collection_name(self, name):
        """Set the name of the embeddings collection for this warehouse"""
        self.meta = self.meta or {}
        self.meta.setdefault("nlp", {})["collection_name"] = name

    def _create_or_update_datasources(
        self, ds_configs, skip_integrity_checks=False, nlp=False
    ):
        """Given a set of datasource configs, create or update the datasources
        on this warehouse. If a datasource exists already it will be updated by
        applying the datasource config. Otherwise this attempts to create a
        datasource from the config.

        **Parameters:**

        * **ds_configs** - (*dict*) A dict mapping datasource names to
        datasource configs
        * **skip_integrity_checks** - (*bool, optional*) If True, skip warehouse
        integrity checks
        * **nlp** - (*bool, optional*) If true, allow NLP analysis when creating
        fields

        """
        for ds_name in ds_configs:
            if ds_name in self.datasource_names:
                self.get_datasource(ds_name).apply_config(ds_configs[ds_name])
                continue

            ds = DataSource(ds_name, config=ds_configs[ds_name], nlp=nlp)
            self.add_datasource(ds, skip_integrity_checks=skip_integrity_checks)

    def _clear_supported_dimension_cache(self):
        """Clear the cache of supported dimensions"""
        self._supported_dimension_cache = {}

    def _check_reserved_field_names(self, adhoc_datasources=None):
        """Integrity check against reserved field names"""
        errors = []
        for field in self.get_field_names(adhoc_fms=adhoc_datasources):
            if field in RESERVED_FIELD_NAMES:
                errors.append("Field name %s is reserved" % field)
        return errors

    def _check_conflicting_fields(self, adhoc_datasources=None):
        """Integrity check for conflicting field definitions"""
        errors = []

        for field in self.get_field_names(adhoc_fms=adhoc_datasources):
            if self.has_metric(
                field, adhoc_fms=adhoc_datasources
            ) and self.has_dimension(field, adhoc_fms=adhoc_datasources):
                errors.append("Field %s is in both metrics and dimensions" % field)

            instances = self.get_field_instances(field, adhoc_fms=adhoc_datasources)

            field_type_mismatch = False
            data_type_mismatch = False
            aggregation_mismatch = False
            field_type = None
            field_aggr = None
            field_is_numeric = None

            for fm, field_def in instances.items():
                if not field_type:
                    field_type = field_def.field_type
                    field_is_numeric = is_numeric_type(field_def.sa_type)
                    field_aggr = getattr(field_def, "aggregation", None)
                    continue

                if field_def.field_type != field_type:
                    field_type_mismatch = True

                if is_numeric_type(field_def.sa_type) != field_is_numeric:
                    data_type_mismatch = True

                if field_def.field_type == FieldTypes.METRIC:
                    if field_def.aggregation != field_aggr:
                        aggregation_mismatch = True

            if field_type_mismatch:
                errors.append("Field %s is in both metrics and dimensions" % field)
            if data_type_mismatch:
                errors.append("Field %s has data type mismatches" % field)
            if aggregation_mismatch:
                errors.append("Field %s has aggregation mismatches" % field)

        return errors

    def _check_fields_have_type(self, adhoc_datasources=None):
        """Integrity check for field types"""
        errors = []

        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            for table in ds.metadata.tables.values():
                if not is_active(table):
                    continue

                for column in table.c:
                    if not is_active(column):
                        continue

                    for field in column.zillion.get_field_names():
                        if not (
                            self.has_metric(field, adhoc_fms=adhoc_datasources)
                            or self.has_dimension(field, adhoc_fms=adhoc_datasources)
                        ):
                            errors.append(
                                "Field %s for column %s->%s is not defined as a metric or dimension"
                                % (field, ds.name, column_fullname(column))
                            )

        return errors

    def _check_primary_key_dimensions(self, adhoc_datasources=None):
        """Integrity check for primary keys"""
        errors = []

        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            for table in ds.metadata.tables.values():
                if not is_active(table):
                    continue

                primary_key = table.zillion.primary_key
                table_dims = get_table_dimensions(
                    self, table, adhoc_fms=adhoc_datasources
                )

                for pk_field in primary_key:
                    if not self.has_dimension(pk_field):
                        errors.append(
                            "Primary key field is not a dimension: %s" % pk_field
                        )
                    if pk_field not in table_dims:
                        errors.append(
                            "Primary key dimension %s is not in table %s"
                            % (pk_field, table.fullname)
                        )

        return errors

    def _check_weighting_metrics(self, adhoc_datasources=None):
        """Integrity check for weighting metrics"""
        errors = []

        for metric in self.get_metrics(adhoc_fms=adhoc_datasources).values():
            if not metric.weighting_metric:
                continue

            for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
                tables = ds.get_tables_with_field(metric.name)
                if not tables:
                    continue

                for table in tables:
                    if not metric.weighting_metric in get_table_fields(table):
                        errors.append(
                            "Table %s->%s has metric %s but not weighting metric %s"
                            % (
                                ds.name,
                                table.fullname,
                                metric.name,
                                metric.weighting_metric,
                            )
                        )

        return errors

    def _check_required_grain(self, adhoc_datasources=None):
        """Integrity check for required_grain settings"""
        errors = []

        for metric in self.get_metrics(adhoc_fms=adhoc_datasources).values():
            if not metric.required_grain:
                continue

            for field in metric.required_grain:
                if not self.has_dimension(field, adhoc_fms=adhoc_datasources):
                    errors.append(
                        "Metric %s references unknown dimension %s in required_grain"
                        % (metric.name, field)
                    )

        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            for table in ds.metadata.tables.values():
                if not is_active(table):
                    continue

                for column in table.c:
                    if not is_active(column):
                        continue
                    if not column.zillion.required_grain:
                        continue

                    for field in column.zillion.required_grain:
                        if not self.has_dimension(field, adhoc_fms=adhoc_datasources):
                            errors.append(
                                "Column %s->%s references unknown dimension %s in required_grain"
                                % (ds.name, column_fullname(column), field)
                            )

        return errors

    def _check_incomplete_dimensions(self, adhoc_datasources=None):
        """Integrity check for incomplete_dimensions settings"""
        errors = []

        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            for table in ds.metadata.tables.values():
                if not is_active(table):
                    continue

                if not table.zillion.incomplete_dimensions:
                    continue

                for field in table.zillion.incomplete_dimensions:
                    if not self.has_dimension(field, adhoc_fms=adhoc_datasources):
                        errors.append(
                            "Table %s->%s references unknown dimension %s in incomplete_dimensions"
                            % (ds.name, table.fullname, field)
                        )

        return errors

    def _check_valid_table_parents(self, adhoc_datasources=None):
        """Integrity check for table parents"""
        errors = []

        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            for table in ds.metadata.tables.values():
                if not is_active(table):
                    continue

                if not table.zillion.parent:
                    continue

                if not ds.has_table(table.zillion.parent, check_active=False):
                    errors.append(
                        "Table %s->%s references unknown parent %s"
                        % (ds.name, table.fullname, table.zillion.parent)
                    )

                parent = ds.get_table(table.zillion.parent, check_active=False)
                if table.zillion.type != parent.zillion.type:
                    errors.append(
                        "Table %s->%s references parent %s with different table type"
                        % (ds.name, table.fullname, table.zillion.parent)
                    )

        return errors

    def _get_supported_dimensions_for_metric(
        self, metric, use_cache=True, adhoc_datasources=None
    ):
        """Get a set of all supported dimensions for a metric

        **Parameters:**

        * **metric** - (*str or Metric*) A metric name or Metric object
        * **use_cache** - (*bool, optional*) If True, try to pull the result
        from the supported dimension cache
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers
        specific to this request

        **Returns:**

        (*set*) - A set of dims supported by this metric

        """
        dims = set()
        metric = self.get_metric(metric, adhoc_fms=adhoc_datasources)

        if use_cache and metric.name in self._supported_dimension_cache:
            return self._supported_dimension_cache[metric]

        for ds in self.datasources:
            ds_tables = ds.get_tables_with_field(metric.name)
            used_tables = set()

            for ds_table in ds_tables:
                if ds_table.fullname not in used_tables:
                    dims |= get_table_dimensions(
                        self, ds_table, adhoc_fms=adhoc_datasources
                    )
                    used_tables.add(ds_table.fullname)

                desc_tables = ds.find_descendent_tables(ds_table)
                for desc_table in desc_tables:
                    if desc_table not in used_tables:
                        dims |= get_table_dimensions(
                            self, ds.get_table(desc_table), adhoc_fms=adhoc_datasources
                        )
                        used_tables.add(desc_table)

        self._supported_dimension_cache[metric] = dims
        return dims

    def _get_supported_dimensions(self, metrics, adhoc_datasources=None):
        """Get all of the supported dimensions shared among these metrics

        **Parameters:**

        * **metrics** - (*list*) A list of metric names or Metric objects
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers
        specific to this request

        **Returns:**

        (*set*) - A set of dims supported by all of these metrics

        """
        dims = set()
        for metric in metrics:
            supported_dims = self._get_supported_dimensions_for_metric(
                metric, adhoc_datasources=adhoc_datasources
            )
            dims = (dims & supported_dims) if len(dims) > 0 else supported_dims
        return dims

    def _get_ds_tables_with_metric(self, metric, adhoc_datasources=None):
        """Get a list of tables in each datasource that provide this metric

        **Parameters:**

        * **metric** - (*str*) The name of a metric
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers
        specific to this request

        **Returns:**

        (*dict*) - A dict mapping datasource names to a list of tables
        supporting this metric

        """
        ds_tables = defaultdict(list)
        count = 0
        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            tables = ds.get_tables_with_field(metric)
            if not tables:
                continue
            ds_tables[ds.name] = tables
            count += 1
        dbg(
            "found %d datasources, %d columns for metric %s"
            % (len(ds_tables), count, metric)
        )
        return ds_tables

    def _get_ds_dim_tables_with_dim(self, dim, adhoc_datasources=None):
        """Get a list of tables in each datasource that provide this dimension

        **Parameters:**

        * **dim** - (*str*) The name of a dimension
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers
        specific to this request

        **Returns:**

        (*dict*) - A dict mapping datasource names to a list of tables
        supporting this dimension

        """
        ds_tables = defaultdict(list)
        count = 0
        for ds in self.get_field_managers(adhoc_fms=adhoc_datasources):
            tables = ds.get_dim_tables_with_dim(dim)
            if not tables:
                continue
            ds_tables[ds.name] = tables
            count += 1
        dbg(
            "found %d datasources, %d columns for dim %s" % (len(ds_tables), count, dim)
        )
        return ds_tables

    def _get_ds_table_sets(
        self, ds_tables, field, grain, dimension_grain, adhoc_datasources=None
    ):
        """Get a list of TableSets that can satisfy the grain in each datasource

        **Parameters:**

        * **ds_tables** - (*dict*) A mapping of datasource names to tables
        containing the field
        * **field** - (*str*) A field name that is contained in the datasource
        tables
        * **grain** - (*list*) A list of dimension names representing the full
        grain required including dimension and criteria grain
        * **dimension_grain** - (*list of str*) A list of dimension names
        representing the requested dimensions for report grouping
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers for
        this request

        **Returns:**

        (*dict*) - A dict mapping datasource names to possible TableSets that
        satisfy the field/grain requirements

        """
        ds_table_sets = {}
        for ds_name, ds_tables_with_field in ds_tables.items():
            ds = self.get_datasource(ds_name, adhoc_datasources=adhoc_datasources)
            possible_table_sets = ds.find_possible_table_sets(
                ds_tables_with_field, field, grain, dimension_grain
            )
            if not possible_table_sets:
                continue
            ds_table_sets[ds_name] = possible_table_sets
        dbg(ds_table_sets)
        return ds_table_sets

    def _choose_best_datasource(self, ds_names):
        """Choose the best datasource to use. Currently this will used the
        ds_priority attribute to make the decision. If that is not defined it
        will just take the first datasource in the list. There is room for
        improvement by taking into account expected query performance.

        **Parameters:**

        * **ds_names** - (*list*) A list of datasource names

        **Returns:**

        (*str*) - The name of the best datasource to use

        """
        if self.ds_priority:
            for ds_name in self.ds_priority:
                if ds_name in ds_names:
                    return ds_name

        info("No datasource priorities established, choosing first option")
        raiseifnot(ds_names, "No datasource names provided")
        return ds_names[0]

    def _choose_best_table_set(self, ds_table_sets):
        """Choose the best TableSet to use among possible options. This will
        first choose the best datasource among the available options, and then
        the best TableSet within that. Currently the best TableSet is chosen
        simply as the one with the fewest number of tables in its join.

        **Parameters:**

        * **ds_table_sets** - (*dict*) A dict mapping datasource names to lists
        of possible TableSets

        **Returns:**

        (*TableSet*) - The best available TableSet

        """
        ds_name = self._choose_best_datasource(list(ds_table_sets.keys()))
        table_sets = ds_table_sets[ds_name]
        best_priority = None
        table_sets_by_priority = defaultdict(list)

        for ts in table_sets:
            priority = ts.ds_table.zillion.priority
            raiseif(
                priority is None, f"Expected priority on table {ts.ds_table.fullname}"
            )
            table_sets_by_priority[priority].append(ts)
            if best_priority is None:
                best_priority = priority
            else:
                best_priority = min(priority, best_priority)

        info(f"Best table set priority = {best_priority}")
        top_table_sets = table_sets_by_priority[best_priority]

        if len(top_table_sets) > 1:
            # TODO: table set priorities based on expected query performance
            info("Picking smallest of %d top table sets" % len(top_table_sets))

        return sorted(top_table_sets, key=len)[0]

    def _generate_unsupported_grain_msg(self, grain, metric, adhoc_datasources=None):
        """Generate a messaged that aims to help pinpoint why a metric can not
        meet a specific grain

        **Parameters:**

        * **grain** - (*list*) A list of dimensions
        * **metric** - (*str*) A metric name
        * **adhoc_datasources** - (*list, optional*) A list of FieldManagers for
        this report

        **Returns:**

        (*str*) - A message explaining the unsupported grain issue

        """
        grain = grain or set()
        supported = self._get_supported_dimensions_for_metric(
            metric, adhoc_datasources=adhoc_datasources
        )
        unsupported = grain - supported
        if unsupported:
            msg = (
                "metric %s can not meet grain %s due to unsupported dimensions: %s"
                % (metric, grain, unsupported)
            )
        else:
            msg = "metric %s can not meet grain %s in any single datasource" % (
                metric,
                grain,
            )
        return msg

    @classmethod
    def load(cls, id):
        """Load a Warehouse from a Warehouse ID

        **Parameters:**

        * **id** - (*int*) A Warehouse ID

        **Returns:**

        (*Warehouse*) - A Warehouse object

        """
        wh = cls._load_warehouse(id)
        if not wh:
            raise InvalidWarehouseIdException(
                "Could not find Warehouse for id: %s" % id
            )

        params = json.loads(wh["params"])
        meta = json.loads(wh["meta"]) if wh["meta"] else None
        result = Warehouse(**params)
        result.meta = meta
        result.name = wh.name
        result.id = id
        return result

    @classmethod
    def load_warehouse_for_report(cls, spec_id):
        """Load the warehouse corresponding to the ReportSpec

        **Parameters:**

        * **spec_id** - (*int*) A ReportSpec ID

        **Returns:**

        (*Warehouse*) - A Warehouse object

        """
        wh_id = Report.load_warehouse_id_for_report(spec_id)
        raiseifnot(wh_id, "No warehouse ID found for spec ID %s" % spec_id)
        return cls.load(wh_id)

    @classmethod
    def load_report_and_warehouse(cls, spec_id):
        """Load a Report and Warehouse from a ReportSpec. The Warehouse
        will be populated on the returned Report object.

        **Parameters:**

        * **spec_id** - (*int*) A ReportSpec ID

        **Returns:**

        (*Report*) - A Report built from this report spec

        """
        wh = cls.load_warehouse_for_report(spec_id)
        return wh.load_report(spec_id)

    @classmethod
    def delete(cls, id):
        """Delete a saved warehouse. Note that this does not delete
        any report specs that reference this warehouse ID.

        **Parameters:**

        * **id** - (*int*) The ID of a Warehouse to delete

        """
        s = Warehouses.delete().where(Warehouses.c.id == id)
        conn = zillion_engine.connect()
        try:
            conn.execute(s)
        finally:
            conn.close()

    @classmethod
    def _load_warehouse(cls, id):
        """Get a Warehouse row from a Warehouse ID

        **Parameters:**

        * **id** - (*int*) The ID of the Warehouse to load

        **Returns:**

        (*dict*) - A Warehouse row

        """
        s = sa.select(Warehouses.c).where(Warehouses.c.id == id)
        conn = zillion_engine.connect()
        try:
            result = conn.execute(s)
            row = result.fetchone()
            return row
        finally:
            conn.close()

    @classmethod
    def from_db_file(cls, *args, **kwargs):
        """Pass args through to DataSource.from_db_file and create a Warehouse"""
        ds = DataSource.from_db_file(*args, **kwargs)
        return cls(datasources=[ds])

    @classmethod
    def from_data_file(cls, *args, **kwargs):
        """Pass args through to DataSource.from_data_file and create a Warehouse"""
        ds = DataSource.from_data_file(*args, **kwargs)
        return cls(datasources=[ds])



from sqlalchemy import func, text
from sqlalchemy.dialects.postgresql import INTERVAL
from sqlalchemy.sql.functions import concat

from zillion.dialects.conversions import DialectDateConversions


def get_interval(n, t):
    return func.cast(concat(n, f" {t}"), INTERVAL)


class PostgreSQLDialectDateConversions(DialectDateConversions):
    @classmethod
    def date_year_start(cls, x):
        return func.TO_DATE(str(x) + "-01-01", "YYYY-MM-DD")

    @classmethod
    def date_year_plus_year(cls, x):
        return func.TO_DATE(str(x) + "-01-01", "YYYY-MM-DD") + get_interval(1, "YEARS")

    @classmethod
    def datetime_year_end(cls, x):
        return (
            func.TO_DATE(str(x) + "-01-01", "YYYY-MM-DD")
            + get_interval(1, "YEARS")
            - get_interval(1, "SECONDS")
        )

    @classmethod
    def date_month_start(cls, x):
        return func.TO_DATE(str(x) + "-01", "YYYY-MM")

    @classmethod
    def date_month_plus_month(cls, x):
        return func.TO_DATE(str(x) + "-01", "YYYY-MM") + get_interval(1, "MONTHS")

    @classmethod
    def datetime_month_end(cls, x):
        return (
            func.TO_DATE(str(x) + "-01", "YYYY-MM")
            + get_interval(1, "MONTHS")
            - get_interval(1, "SECONDS")
        )

    @classmethod
    def date_plus_day(cls, x):
        return func.TO_DATE(x, "YYYY-MM-DD") + get_interval(1, "DAYS")

    @classmethod
    def datetime_day_end(cls, x):
        return (
            func.TO_DATE(x, "YYYY-MM-DD")
            + get_interval(1, "DAYS")
            - get_interval(1, "SECONDS")
        )

    @classmethod
    def datetime_hour_plus_hour(cls, x):
        return func.TO_TIMESTAMP(x, "YYYY-MM-DD HH24:MI:SS") + get_interval(1, "HOURS")

    @classmethod
    def datetime_hour_end(cls, x):
        return (
            func.TO_TIMESTAMP(x, "YYYY-MM-DD HH24:MI:SS")
            + get_interval(1, "HOURS")
            - get_interval(1, "SECONDS")
        )

    @classmethod
    def datetime_minute_plus_minute(cls, x):
        return func.TO_TIMESTAMP(x, "YYYY-MM-DD HH24:MI:SS") + get_interval(
            1, "MINUTES"
        )

    @classmethod
    def datetime_minute_end(cls, x):
        return (
            func.TO_TIMESTAMP(x, "YYYY-MM-DD HH24:MI:SS")
            + get_interval(1, "MINUTES")
            - get_interval(1, "SECONDS")
        )


POSTGRESQL_DIALECT_CONVERSIONS = {
    "year": {
        "ds_formula": "EXTRACT(YEAR FROM {})",
        "ds_criteria_conversions": PostgreSQLDialectDateConversions.get_year_criteria_conversions(),
    },
    "quarter": "TO_CHAR({}, 'FMYYYY-\"Q\"Q')",
    "quarter_of_year": "EXTRACT(QUARTER FROM {})",
    "month": {
        "ds_formula": "TO_CHAR({}, 'FMYYYY-MM')",
        "ds_criteria_conversions": PostgreSQLDialectDateConversions.get_month_criteria_conversions(),
    },
    "month_name": "TO_CHAR({}, 'FMMonth')",
    "month_of_year": "EXTRACT(MONTH FROM {})",
    "week_of_month": "EXTRACT(WEEK FROM {}) - EXTRACT(WEEK FROM CAST(DATE_TRUNC('month', {}) as date)) + 1",
    "week_of_year": "EXTRACT(WEEK FROM {})",  # HACK: attempt to get tests compatible with mysql and sqlite
    "period_of_month_7d": "FLOOR((EXTRACT(DAY FROM {}) - 1) / 7) + 1",
    "date": {
        "ds_formula": "TO_CHAR({}, 'FMYYYY-MM-DD')",
        "ds_criteria_conversions": PostgreSQLDialectDateConversions.get_date_criteria_conversions(),
    },
    "day_name": "TO_CHAR({}, 'FMDay')",
    "day_of_week": "EXTRACT(ISODOW FROM {})",  # Monday = 1
    "is_weekday": (
        "CASE EXTRACT(ISODOW FROM {}) "
        "WHEN 1 THEN 1 "
        "WHEN 2 THEN 1 "
        "WHEN 3 THEN 1 "
        "WHEN 4 THEN 1 "
        "WHEN 5 THEN 1 "
        "WHEN 6 THEN 0 "
        "WHEN 7 THEN 0 "
        "ELSE NULL "
        "END"
    ),
    "day_of_month": "EXTRACT(DAY FROM {})",
    "day_of_year": "EXTRACT(DOY FROM {})",
    "hour": {
        "ds_formula": "TO_CHAR({}, 'FMYYYY-MM-DD HH24:00:00')",
        "ds_criteria_conversions": PostgreSQLDialectDateConversions.get_hour_criteria_conversions(),
    },
    "hour_of_day": "EXTRACT(HOUR FROM {})",
    "minute": {
        "ds_formula": "TO_CHAR({}, 'FMYYYY-MM-DD HH24:MI:00')",
        "ds_criteria_conversions": PostgreSQLDialectDateConversions.get_minute_criteria_conversions(),
    },
    "minute_of_hour": "EXTRACT(MINUTE FROM {})",
    "datetime": {
        "ds_formula": "TO_CHAR({}, 'FMYYYY-MM-DD HH24:MI:SS')",
        "ds_criteria_conversions": PostgreSQLDialectDateConversions.get_datetime_criteria_conversions(),
    },
    "unixtime": "EXTRACT(epoch from {})",
}



from datetime import datetime, timedelta
import hashlib
import re
import struct
import time

try:
    from qdrant_client import QdrantClient
    from qdrant_client.http import models as rest
    from qdrant_client.http.models import Distance, VectorParams
    from langchain.chains import LLMChain
    from langchain.embeddings.openai import OpenAIEmbeddings
    from langchain.llms import OpenAI, OpenAIChat
    from langchain.prompts import PromptTemplate
    from langchain.vectorstores import Qdrant
    from pydantic import Extra
except ImportError:
    pass

from sqlalchemy import and_
from sqlalchemy.schema import CreateTable
from tlbx import raiseifnot, st, json, rgetkey

from zillion.core import (
    dbg,
    info,
    warn,
    error,
    zillion_config,
    nlp_installed,
    RollupTypes,
    FieldTypes,
)

LLM_MAX_TOKENS = -1
LLM_REQUEST_TIMEOUT = 20
OPENAI_DAVINCI_MODEL_NAME = "text-davinci-003"
OPENAI_VECTOR_SIZE = 1536
MIN_FIELD_SIMILARITY_SCORE = 0.8
DEFAULT_VECTOR_SIZE = OPENAI_VECTOR_SIZE
DEFAULT_WAREHOUSE_COLLECTION_NAME = "default_warehouse_fields"

embeddings_api = None


def hash_text(text):
    """Hash a string to a 32-character hex string""" ""
    return hashlib.md5(text.encode("utf-8")).hexdigest()


if nlp_installed:
    from zillion.model import (
        EmbeddingsCache as EmbeddingsCacheTable,
        zillion_engine,
    )

    class EmbeddingsCache:
        """
        A cache for embeddings. This is a wrapper around a database table
        that stores embeddings for text. It also provides a cache for
        embeddings that have been retrieved from the database.

        **Parameters:**

        * **table** - (*SQLAlchemy Table*) The database table to use for the cache
        * **model** - (*str*) The model to use for the embeddings
        * **binary** - (*bool, optional) Whether embeddings are store in binary format
        * **binary_size** - (*int, optional) The size of the vector for converting to/from binary

        """

        def __init__(
            self,
            table,
            model,
            binary=True,
            binary_size=DEFAULT_VECTOR_SIZE,
        ):
            self.table = table
            self.model = model
            self.binary = binary
            self.binary_size = binary_size
            self.conn = zillion_engine.connect()
            self.init_cache()

        @classmethod
        def get_text_hash(cls, text):
            """Get the hash of a text string"""
            return hash_text(text)

        @property
        def cache(self):
            if not hasattr(self, "_cache"):
                self._cache = {}
            return self._cache

        @cache.setter
        def cache(self, value):
            self._cache = value

        def decode(self, blob):
            """Decode a binary blob into a list of floats"""
            return struct.unpack("f" * self.binary_size, blob)

        def encode(self, values):
            """Encode a list of floats into a binary blob"""
            return struct.pack("f" * self.binary_size, *values)

        def init_cache(self):
            """Initialize the in-memory cache"""
            stmt = self.table.select()
            result = self.conn.execute(stmt).fetchall()
            for row in result:
                text_hash = row["text_hash"]
                vector = row["vector"]
                if self.binary:
                    vector = self.decode(vector)
                self.cache[(text_hash, self.model)] = vector
            info(f"Initialized embeddings cache with {len(self.cache)} records")

        def _get_key(self, text):
            """Get the in-memory cache key. Matches the unique/primary key of
            the cache database table"""
            return (self.get_text_hash(text), self.model)

        def __getitem__(self, key):
            """Get an embedding from the cache. If it's not in the cache,
            it will be retrieved from the database.

            **Parameters:**

            * **key** - (*str*) The text to get the embedding for

            **Returns:**

            * (*list*) The embedding vector

            """
            cache_key = self._get_key(key)
            text_hash, _ = cache_key
            value = self.cache.get(cache_key)
            if value is None:
                stmt = self.table.select().where(
                    and_(
                        self.table.c.text_hash == text_hash,
                        self.table.c.model == self.model,
                    )
                )
                result = self.conn.execute(stmt).fetchone()
                if result:
                    value = result["vector"]
                    if self.binary:
                        value = self.decode(value)
                    self.cache[cache_key] = value
            return value

        def __setitem__(self, key, value):
            """Set an embedding in the cache. If it's not in the cache,
            it will be added to the database.

            **Parameters:**

            * **key** - (*str*) The text to set the embedding for
            * **value** - (*list*) The embedding vector

            """
            record = self.__getitem__(key)
            cache_key = self._get_key(key)
            text_hash, _ = cache_key
            if self.binary:
                vector = self.encode(value)

            if record:
                stmt = (
                    self.table.update()
                    .where(
                        and_(
                            self.table.c.text_hash == text_hash,
                            self.table.c.model == self.model,
                        )
                    )
                    .values(vector=vector, text=key)
                )
            else:
                stmt = self.table.insert().values(
                    text_hash=text_hash, model=self.model, text=key, vector=vector
                )

            self.conn.execute(stmt)
            self.cache[cache_key] = value

        def __delitem__(self, key):
            """Delete an embedding from the cache

            **Parameters:**

            * **key** - (*str*) The text to delete the embedding for

            """
            cache_key = self._get_key(key)
            text_hash, _ = cache_key
            stmt = self.table.delete().where(self.table.c.text_hash == text_hash)
            self.conn.execute(stmt)
            if cache_key in self.cache:
                del self.cache[cache_key]

    class OpenAIEmbeddingsCached(OpenAIEmbeddings):
        """OpenAI Embeddings with a cache for faster retrieval and
        to avoid extra API calls"""

        class Config:
            """Configuration for this pydantic object."""

            extra = Extra.allow

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._cache = EmbeddingsCache(
                EmbeddingsCacheTable,
                self.query_model_name,
                binary=True,
                binary_size=OPENAI_VECTOR_SIZE,
            )

        def embed_query(self, query):
            """Embed a query and cache the result"""
            res = self._cache[query]
            if res:
                return res
            embedding = super().embed_query(query)
            self._cache[query] = embedding
            return embedding

        def embed_documents(self, documents):
            """Embed documents and cache the results"""
            uncached = [d for d in documents if not self._cache[d]]
            info(f"Embedding {len(uncached)}/{len(documents)} uncached documents")
            if uncached:
                embeddings = super().embed_documents(uncached)
                for document, embedding in zip(uncached, embeddings):
                    self._cache[document] = embedding
            return [self._cache[d] for d in documents]

    class QdrantCustom(Qdrant):
        """Qdrant with custom ID generation and bulk embedding"""

        @classmethod
        def get_id(cls, text):
            """Get the hash of a text string to use as an ID"""
            return hash_text(text)

        def add_texts(self, texts, metadatas=None, bulk_embedder=None):
            """Add texts to Qdrant. If a bulk embedder is provided, it will be
            used to embed the texts in bulk. Otherwise, the texts will be
            embedded one at a time.

            **Parameters:**

            * **texts** - (*list*) List of texts to add to Qdrant
            * **metadatas** - (*list, optional*) List of metadata dictionaries to add to Qdrant
            * **bulk_embedder - (*callable, optional*) A function that takes a list of texts and
            returns a list of embeddings

            **Returns:**

            * **ids** - (*list*) List of IDs for the texts added to Qdrant

            """
            ids = [self.get_id(text) for text in texts]
            if bulk_embedder:
                vectors = bulk_embedder(texts)
            else:
                vectors = [self.embedding_function(text) for text in texts]
            payloads = self._build_payloads(
                texts,
                metadatas,
                self.content_payload_key,
                self.metadata_payload_key,
            )

            start = time.time()
            self.client.upsert(
                collection_name=self.collection_name,
                points=rest.Batch.construct(
                    ids=ids, vectors=[list(v) for v in vectors], payloads=payloads
                ),
            )
            info(f"Added {len(texts)} texts to Qdrant in {time.time() - start:.2f}s")
            return ids

        def similarity_search_with_score(self, query, k=4, **kwargs):
            """Search for similar texts to a query and return the results with scores

            **Parameters:**

            * **query** - (*str*) The query text to search for
            * **k** - (*int, optional*) The number of results to return
            * **kwargs** - (*dict, optional*) Additional keyword arguments to pass to Qdrant

            **Returns:**

            * **results** - (*list*) List of tuples of the form (document, score)

            """
            embedding = self.embedding_function(query)
            results = self.client.search(
                collection_name=self.collection_name,
                query_vector=list(embedding),
                with_payload=True,
                limit=k,
                **kwargs,
            )
            return [
                (
                    self._document_from_scored_point(
                        result, self.content_payload_key, self.metadata_payload_key
                    ),
                    result.score,
                )
                for result in results
            ]

    class EmbeddingsAPI:
        """API for embedding texts and querying Qdrant"""

        def __init__(self):
            key = zillion_config["OPENAI_API_KEY"]
            self.embeddings = OpenAIEmbeddingsCached(openai_api_key=key)
            # We delay connecting to Qdrant until we need it. This way
            # we can init the API globally without Qdrant necessarily running.
            self.client = None

        def connect(self):
            """Connect to Qdrant"""
            host = zillion_config["QDRANT_HOST"]
            info(f"Connecting to Qdrant host: {host}...")
            if host == ":memory:":
                self.client = QdrantClient(location=host)
            elif host.startswith("http") or host in [
                "localhost",
                "127.0.0.1",
                "host.docker.internal",
                "qdrant",  # Docker container name
            ]:
                self.client = QdrantClient(host=host, port=6333, prefer_grpc=True)
            else:
                self.client = QdrantClient(path=host)

        def ensure_client(self):
            """Make sure we have a client. If not, connect to Qdrant."""
            if not self.client:
                self.connect()

        def embed_documents(self, rows):
            """Embed a list of texts"""
            return self.embeddings.embed_documents(rows)

        def embed_query(self, query):
            """Embed a query"""
            return self.embeddings.embed_query(query)

        def recreate_collection(
            self,
            collection_name,
            vector_size=None,
            distance=Distance.COSINE,
            sample=None,
            **kwargs,
        ):
            """Create or recreate a collection in Qdrant

            **Parameters:**

            * **collection_name** - (*str*) Name of the collection
            * **vector_size** - (*int, optional*) Size of the vector. If not provided, will be inferred
            from the sample.
            * **distance** - (*Distance, optional*) Distance metric to use. Defaults to cosine.
            * **sample** - (*str, optional*) A sample text to use to infer the vector size.

            **Returns:**

            * **collection** - (*QdrantCollection*) The collection

            """
            info(f"Recreating collection {collection_name}...")
            if sample and vector_size is None:
                vector_size = len(self.embed_documents([sample])[0])
            elif vector_size is None:
                vector_size = DEFAULT_VECTOR_SIZE

            self.ensure_client()
            self.client.recreate_collection(
                collection_name=collection_name,
                vectors_config=VectorParams(size=vector_size, distance=distance),
                **kwargs,
            )
            return self.get_collection(collection_name)

        def create_collection_if_necessary(
            self,
            collection_name,
            vector_size=None,
            distance=Distance.COSINE,
            sample=None,
            **kwargs,
        ):
            """Create a collection if it doesn't already exist. If it does exist,
            just return the collection.

            **Parameters:**

            * **collection_name** - (*str*) Name of the collection
            * **vector_size** - (*int, optional*) Size of the vector. If not provided, will be inferred
            from the sample.
            * **distance** - (*Distance, optional*) Distance metric to use. Defaults to cosine.
            * **sample** - (*str, optional*) A sample text to use to infer the vector size.

            **Returns:**

            * **collection** - (*QdrantCollection*) The collection

            """
            try:
                collection = self.get_collection(collection_name)
                if collection:
                    info(
                        f"Collection {collection_name} already exists. Skipping creation."
                    )
                    return collection
            except Exception as e:
                if "not found" not in str(e).lower():
                    raise e

            return self.recreate_collection(
                collection_name,
                vector_size=vector_size,
                distance=distance,
                sample=sample,
                **kwargs,
            )

        def add_texts(
            self, collection_name, texts, metadatas=None, force_recreate=False
        ):
            """Add texts to Qdrant. See QdrantCustom.add_texts for details."""
            self.ensure_client()
            if force_recreate:
                self.recreate_collection(
                    collection_name, sample=texts[0], metadatas=metadatas
                )
            else:
                self.create_collection_if_necessary(collection_name, sample=texts[0])

            qdrant = QdrantCustom(
                self.client, collection_name, self.embeddings.embed_query
            )
            qdrant.add_texts(
                texts,
                metadatas=metadatas,
                bulk_embedder=self.embeddings.embed_documents,
            )

        def similarity_search_with_score(self, collection_name, query, **kwargs):
            """Search for similar texts to a query and return the results with scores.
            See QdrantCustom.similarity_search_with_score for details."""
            self.ensure_client()
            qdrant = QdrantCustom(
                self.client, collection_name, self.embeddings.embed_query
            )
            return qdrant.similarity_search_with_score(query, **kwargs)

        def get_collection(self, name):
            """Get a collection by name"""
            self.ensure_client()
            return self.client.get_collection(collection_name=name)

        def delete_collection(self, name):
            """Delete a collection by name"""
            self.ensure_client()
            return self.client.delete_collection(collection_name=name)

        def get_embeddings(
            self, collection_name, with_payload=True, with_vectors=False
        ):
            """Get all embeddings in a collection

            **Parameters:**

            * **collection_name** - (*str*) Name of the collection
            * **with_payload** - (*bool, optional*) Whether to include the payload. Defaults to True.
            * **with_vectors** - (*bool, optional*) Whether to include the vectors. Defaults to False.

            **Returns:**

            * **result** - (*list*) List of embeddings

            """
            self.ensure_client()
            collection = self.get_collection(collection_name)
            raiseifnot(collection, f"No collection found: {collection_name}")
            result = []
            offset = None
            while True:
                points, next_page_offset = self.client.scroll(
                    collection_name=collection_name,
                    with_payload=with_payload,
                    with_vectors=with_vectors,
                    limit=500,
                    offset=offset,
                )
                result.extend([dict(x) for x in points])
                if next_page_offset is None:
                    break
                offset = next_page_offset
            return result

        def delete_embeddings(self, collection_name, texts):
            """Delete embeddings by text

            **Parameters:**

            * **collection_name** - (*str*) Name of the collection
            * **texts** - (*list*) List of texts to delete

            """
            ids = [QdrantCustom.get_id(text) for text in texts]
            self.ensure_client()
            self.client.delete(
                collection_name=collection_name,
                points_selector=rest.PointIdsList(points=ids),
            )

        def upsert_embedding(self, collection_name, text, payload):
            """Upsert an embedding

            **Parameters:**

            * **collection_name** - (*str*) Name of the collection
            * **text** - (*str*) Text to upsert
            * **payload** - (*dict*) Payload to upsert

            **Returns:**

            * **result** - (*dict*) Result of the upsert

            """
            self.ensure_client()
            qdrant = QdrantCustom(
                self.client, collection_name, self.embeddings.embed_query
            )
            return qdrant.add_texts([text], metadatas=[payload])


if nlp_installed:
    embeddings_api = EmbeddingsAPI()


def field_name_to_embedding_text(name):
    """Convert a field name to a format for embedding"""
    return name.replace("_", " ").lower()


def get_warehouse_collection_name(warehouse):
    """Get the collection name for a warehouse's embeddings. If it is set on
    the config, use that. Otherwise try to use a warehouse name or fall back
    to a default name."""
    meta = warehouse.meta or {}
    if meta.get("nlp", {}).get("collection_name", None):
        return meta["nlp"]["collection_name"]
    if not warehouse.name:
        warn(
            f"Warehouse has no name. Using default embeddings collection name: {DEFAULT_WAREHOUSE_COLLECTION_NAME}"
        )
        return DEFAULT_WAREHOUSE_COLLECTION_NAME
    return warehouse.name


def warehouse_field_nlp_enabled(warehouse, field_def):
    """Check if NLP is enabled for a field. If it is disabled at any level,
    it will be considered disabled."""

    wh_meta = warehouse.meta or {}
    field_meta = field_def.meta or {}

    # 1) Check disabled groups at warehouse level
    if rgetkey(wh_meta, "nlp.field_disabled_groups", None) and "group" in field_meta:
        if field_meta["group"] in wh_meta["nlp"]["field_disabled_groups"]:
            return False

    # 2) Check disabled fields by regex at warehouse level
    if rgetkey(wh_meta, "nlp.field_disabled_patterns", None):
        for pattern in wh_meta["nlp"]["field_disabled_patterns"]:
            if re.match(pattern, field_def.name):
                return False

    # 3) Check field level settings
    return rgetkey(field_meta, "nlp.enabled", True) is not False


def init_warehouse_embeddings(warehouse, force_recreate=False):
    """
    Initialize embeddings for the warehouse.

    **Parameters:**

    * **warehouse** - (Warehouse) The warehouse to initialize embeddings for.
    * **force_recreate** - (*bool, optional*) If True, force the embeddings
    collection to be recreated from scratch.

    **Returns:**

    (*Embeddings*) - The initialized EmbeddingsAPI object.

    """
    collection_name = get_warehouse_collection_name(warehouse)
    fields = warehouse.get_fields()

    texts = []
    metadatas = []
    count = 0
    for name, fdef in fields.items():
        if not warehouse_field_nlp_enabled(warehouse, fdef):
            continue

        count += 1
        settings = (fdef.meta or {}).get("nlp", {}) or {}

        if settings.get("embedding_text", None):
            # Allow overriding the default embedding text
            emb_texts = settings["embedding_text"]
            if isinstance(emb_texts, str):
                emb_texts = [emb_texts]
        else:
            emb_texts = [field_name_to_embedding_text(name)]

        for emb_text in emb_texts:
            texts.append(emb_text)
            metadatas.append({"name": name, "field_type": fdef.field_type})

    start = time.time()
    info(
        f"Initializing {count}/{len(fields)} fields in embedding collection {collection_name}..."
    )
    embeddings_api.add_texts(
        collection_name=collection_name,
        texts=texts,
        metadatas=metadatas,
        force_recreate=force_recreate,
    )
    info(f"Done in {time.time() - start:3f} seconds.")
    return collection_name


def get_openai_class(model=None):
    """Get the OpenAI class to use for a given model."""
    model = model or zillion_config["OPENAI_MODEL"]
    return OpenAI if model == OPENAI_DAVINCI_MODEL_NAME else OpenAIChat


def get_openai_model_context_size(model):
    """Logic copied from langchain since no util exposed"""
    if model == "text-davinci-003":
        return 4097
    elif model == "text-curie-001":
        return 2048
    elif model == "text-babbage-001":
        return 2048
    elif model == "text-ada-001":
        return 2048
    elif model == "code-davinci-002":
        return 8000
    elif model == "code-cushman-001":
        return 2048
    else:
        return 4097


def build_llm(model=None, max_tokens=None, request_timeout=LLM_REQUEST_TIMEOUT):
    """Build an LLM using langchain and the OpenAI API.

    **Parameters:**

    * **model** - (str, optional) The OpenAI model to use. Defaults to the model specified in the zillion config.
    * **max_tokens** - (int) The maximum number of tokens to generate.
    * **request_timeout** - (int) The maximum number of seconds to wait for a response from the OpenAI API.

    **Returns:**

    (*llm*) - A langchain OpenAI LLM

    """
    model = model or zillion_config["OPENAI_MODEL"]
    key = zillion_config["OPENAI_API_KEY"]
    raiseifnot(model and key, "Missing OpenAI API key or model name in zillion config")
    max_tokens = max_tokens or LLM_MAX_TOKENS
    dbg(f"Building OpenAI {model} chain with max_tokens={max_tokens}")
    openai_class = get_openai_class(model)
    return openai_class(
        model_name=model,
        temperature=0,
        max_tokens=max_tokens,
        request_timeout=request_timeout,
        max_retries=1,
        openai_api_key=key,
    )


def build_chain(
    prompt,
    model=None,
    max_tokens=LLM_MAX_TOKENS,
    request_timeout=LLM_REQUEST_TIMEOUT,
    llm=None,
):
    """Build a chain using langchain and the OpenAI API.

    **Parameters:**

    * **prompt** - (PromptTemplate) The prompt to use.
    * **model** - (str, optional) The OpenAI model to use. Defaults to the model specified in the zillion config.
    * **max_tokens** - (int) The maximum number of tokens to generate.
    * **request_timeout** - (int) The maximum number of seconds to wait for a response from the OpenAI API.
    * **llm** - (llm, optional) The LLM to use. Defaults to a new LLM built using the OpenAI API.

    **Returns:**

    (*LLMChain*) - A langchain LLMChain object.

    """
    llm = llm or build_llm(
        model=model, max_tokens=max_tokens, request_timeout=request_timeout
    )
    return LLMChain(llm=llm, prompt=prompt)


TEXT_TO_REPORT_NO_FIELDS = """You are an expert SQL analyst that takes natural language input and outputs metrics, dimensions, criteria, ordering, and limit settings in JSON format.
If a relative date is specified, such as "yesterday", replace it with the actual date. The current date is: {current_date}

Generic Example 1:

Input: revenue and sales by date for the last 30 days. Rows with more than 5 sales.
Output:
{{
  "metrics": ["revenue", "sales"],
  "dimensions": ["date"],
  "criteria": [
    ["date", ">=", "{thirty_days_ago}"]
  ],
  "row_filters": [
    ["sales", ">", 5]
]
}}

Generic Example 2:

Input: show me the top 10 campaigns by revenue yesterday for ad engine Google. Include totals.
Output:
{{
  "metrics": ["revenue"],
  "dimensions": ["campaign"],
  "criteria": [
    ["date", "=", "{yesterday}"],
    ["ad_engine", "=", "Google"]
  ],
  "limit": 10,
  "order_by": [
    ["revenue", "desc"]
  ],
  "rollup": "totals"
}}

----
Complete the following. Use JSON format and include no other commentary.
Input: {query}
JSON Output:"""


TEXT_TO_REPORT_ALL_FIELDS = """You are an expert SQL analyst that takes natural language input and outputs metrics, dimensions, criteria, ordering, and limit settings in JSON format.
If a relative date is specified, such as "yesterday", replace it with the actual date. The current date is: {current_date}

Supported metrics:
{metrics}

Supported dimensions:
{dimensions}

Generic Example 1:

Input: revenue and sales by date for the last 30 days. Rows with more than 5 sales.
JSON Output:
{{
  "metrics": ["revenue", "sales"],
  "dimensions": ["date"],
  "criteria": [
    ["date", ">=", "{thirty_days_ago}"]
  ],
  "row_filters": [
    ["sales", ">", 5]
]
}}

Generic Example 2:

Input: show me the top 10 campaigns by revenue yesterday for ad engine Google. Include totals.
JSON Output:
{{
  "metrics": ["revenue"],
  "dimensions": ["campaign"],
  "criteria": [
    ["date", "=", "{yesterday}"],
    ["ad_engine", "=", "Google"]
  ],
  "limit": 10,
  "order_by": [
    ["revenue", "desc"]
  ],
  "rollup": "totals"
}}

----
Complete the following. Use JSON format and include no other commentary. Use only supported metrics and dimensions.
Input: {query}
JSON Output:"""

TEXT_TO_REPORT_DIMENSION_FIELDS = """You are an expert SQL analyst that takes natural language input and outputs metrics, dimensions, criteria, ordering, and limit settings in JSON format.
If a relative date is specified, such as "yesterday", replace it with the actual date. The current date is: {current_date}

Supported dimensions:
{dimensions}

Generic Example 1:

Input: revenue and sales by date for the last 30 days. Rows with more than 5 sales.
JSON Output:
{{
  "metrics": ["revenue", "sales"],
  "dimensions": ["date"],
  "criteria": [
    ["date", ">=", "{thirty_days_ago}"]
  ],
  "row_filters": [
    ["sales", ">", 5]
]
}}

Generic Example 2:

Input: show me the top 10 campaigns by revenue yesterday for ad engine Google. Include totals.
JSON Output:
{{
  "metrics": ["revenue"],
  "dimensions": ["campaign"],
  "criteria": [
    ["date", "=", "{yesterday}"],
    ["ad_engine", "=", "Google"]
  ],
  "limit": 10,
  "order_by": [
    ["revenue", "desc"]
  ],
  "rollup": "totals"
}}

----
Complete the following. Use JSON format and include no other commentary.
Input: {query}
JSON Output:"""


def parse_text_to_report_json_output(output):
    """Parse the output of a chain that produces Zillion args as JSON

    **Parameters:**

    * **output** - (str) The output of the chain, expected to be valid JSON.

    **Returns:**

    (*dict*) - A dict of Zillion report params. Any empty values will be removed.

    """
    if not output:
        return None
    try:
        return {k: v for k, v in json.loads(output).items() if v}
    except Exception as e:
        error(f"Error parsing JSON:\n{output}")
        raise


# TODO - share code for configs below

PROMPT_CONFIGS = dict(
    no_fields=dict(
        prompt_text=TEXT_TO_REPORT_NO_FIELDS,
        input_variables=["query", "current_date", "yesterday", "thirty_days_ago"],
        context_func=lambda wh: dict(
            current_date=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            yesterday=str(datetime.now().date() - timedelta(days=1)),
            thirty_days_ago=str(datetime.now().date() - timedelta(days=30)),
        ),
        parser=parse_text_to_report_json_output,
    ),
    all_fields=dict(
        prompt_text=TEXT_TO_REPORT_ALL_FIELDS,
        input_variables=[
            "query",
            "current_date",
            "yesterday",
            "thirty_days_ago",
            "metrics",
            "dimensions",
        ],
        context_func=lambda wh: dict(
            metrics=get_metrics_prompt_str(wh),
            dimensions=get_dimensions_prompt_str(wh),
            current_date=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            yesterday=str(datetime.now().date() - timedelta(days=1)),
            thirty_days_ago=str(datetime.now().date() - timedelta(days=30)),
        ),
        parser=parse_text_to_report_json_output,
    ),
    dimension_fields=dict(
        prompt_text=TEXT_TO_REPORT_DIMENSION_FIELDS,
        input_variables=[
            "query",
            "current_date",
            "yesterday",
            "thirty_days_ago",
            "dimensions",
        ],
        context_func=lambda wh: dict(
            dimensions=get_dimensions_prompt_str(wh),
            current_date=datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            yesterday=str(datetime.now().date() - timedelta(days=1)),
            thirty_days_ago=str(datetime.now().date() - timedelta(days=30)),
        ),
        parser=parse_text_to_report_json_output,
    ),
)


def get_field_name_variants(name):
    """Get a list of possible close variants of a field name. This is
    an 80/20 hack to help with fuzzy matching of field names.

    **Parameters:**

    * **name** - (str) The name of the field

    **Returns:**

    (*set*) - A set of possible variants of the field name
    """
    res = {name}
    for alt in [
        name.lower(),
        name.title(),
        name.replace(" ", "_"),
        name.replace("_", " "),
    ]:
        if alt not in res:
            res.add(alt)
    return res


# NOTE:
# If we ever support fuzzy matching to known dimension values,
# could use a library like: https://github.com/seatgeek/thefuzz


def get_field_fuzzy(warehouse, name, field_type=None):
    """Try to do a fuzzy match to warehouse fields. The field names
    from the LLM are not always exact matches to the warehouse field.

    **Parameters:**

    * **warehouse** - (Warehouse) The warehouse to look for a field match
    * **name** - (str) The name of the field to match
    * **field_type** - (str, optional) The type of field to match. If not
    passed, will match any field type.

    **Returns:**

    (*str*) - The name of the field that was matched, or the original name
    """

    has_field_func = warehouse.has_field
    if field_type == FieldTypes.METRIC:
        has_field_func = warehouse.has_metric
    if field_type == FieldTypes.DIMENSION:
        has_field_func = warehouse.has_dimension

    # Check ~exact matches
    alts = get_field_name_variants(name)
    for alt in alts:
        if has_field_func(alt) and warehouse_field_nlp_enabled(
            warehouse, warehouse.get_field(alt)
        ):
            return alt

    # Check embeddings
    collection_name = warehouse._get_embeddings_collection_name()
    raiseifnot(collection_name, "No embeddings collection name found")
    res = embeddings_api.similarity_search_with_score(collection_name, name)
    if field_type:
        res = [r for r in (res or []) if r[0].metadata["field_type"] == field_type]

    if not res:
        return name

    best, score = res[0]
    if score >= MIN_FIELD_SIMILARITY_SCORE:
        info(f"Found fuzzy match for {name}: {best.metadata['name']} / {score}")
        return best.metadata["name"]
    else:
        warn(f"No good match found for '{name}': {res}")

    # No good match found but guess the original name
    return name


def map_warehouse_report_params(warehouse, report):
    """Map an LLM report params dict to the warehouse's field names.

    **Parameters:**

    * **warehouse** - (Warehouse) The warehouse to map the report params to
    * **report** - (dict) The report params dict

    **Returns:**

    (*dict*) - A new report params dict with the warehouse's field names

    """
    res = {}
    for k, v in report.items():
        if k == "metrics":
            res[k] = [
                get_field_fuzzy(warehouse, m, field_type=FieldTypes.METRIC) for m in v
            ]
        elif k == "dimensions":
            res[k] = [
                get_field_fuzzy(warehouse, d, field_type=FieldTypes.DIMENSION)
                for d in v
            ]
        elif k == "criteria":
            res[k] = [
                (
                    get_field_fuzzy(warehouse, c, field_type=FieldTypes.DIMENSION),
                    op,
                    val,
                )
                for c, op, val in v
            ]
        elif k == "row_filters":
            res[k] = [(get_field_fuzzy(warehouse, f), op, val) for f, op, val in v]
        elif k == "rollup":
            # TODO should we semantic match to valid values?
            # i.e. "summary" -> "totals
            if v and v.lower() in ["totals", "summary"]:
                res[k] = RollupTypes.TOTALS
            elif v and v.lower() == "all":
                res[k] = RollupTypes.ALL
        elif k == "order_by":
            res[k] = [(get_field_fuzzy(warehouse, f), d) for f, d in v]
        elif k == "limit":
            res[k] = int(v) if v is not None else None
        else:
            raise ValueError(f"Unexpected key {k}")
    return res


def get_fields_prompt_str(warehouse, fields):
    """Get a string representing names/types of given fields"""
    res = []
    for name, fdef in fields.items():
        if not warehouse_field_nlp_enabled(warehouse, fdef):
            continue
        if getattr(fdef, "formula", None):
            type_str = "numeric"
        else:
            type_str = str(fdef.type).lower() if fdef.type else "unknown"
        res.append(f"{name} ({type_str})")
    return "\n".join(res)


def get_metrics_prompt_str(warehouse):
    return get_fields_prompt_str(warehouse, warehouse.get_metrics())


def get_dimensions_prompt_str(warehouse):
    return get_fields_prompt_str(warehouse, warehouse.get_dimensions())


class MaxTokensException(Exception):
    pass


def text_to_report_params(query, warehouse=None, prompt_version="no_fields"):
    """Convert a natural language input to Zillion report params

    **Parameters:**

    * **query** - (str) The natural language query to convert.
    * **warehouse** - (Warehouse, optional) The warehouse to map the report params to
    * **prompt_version** - (str) The version of the prompt to use

    **Returns:**

    (*dict*) - A dict of Zillion report params. The field names extracted are
    not guaranteed to exist in any warehouse or datasource and need to be
    analyzed separately.

    """
    prompt_config = PROMPT_CONFIGS[prompt_version]
    context = prompt_config["context_func"](warehouse)
    context["query"] = query
    prompt = PromptTemplate(
        input_variables=prompt_config["input_variables"],
        template=prompt_config["prompt_text"],
    )

    llm = build_llm(max_tokens=-1)
    text_to_report_chain = build_chain(prompt, llm=llm)

    prompt_tokens = text_to_report_chain.llm.get_num_tokens(prompt.format(**context))
    max_tokens = get_openai_model_context_size(text_to_report_chain.llm.model_name)
    if (prompt_tokens + 500) >= max_tokens:
        raise MaxTokensException(f"Prompt is too long: {prompt_tokens} tokens")

    llm_start = time.time()
    output = text_to_report_chain.run(**context).strip(". -\n\r")
    info(f"LLM took {time.time() - llm_start:.3f}s")
    return prompt_config["parser"](output)


NLP_TABLE_RELATIONSHIP_PROMPT = """Given the following tables, what are the suggested foreign key relationships starting from these tables?
Rules:
- If there isn't a good option you must skip that table and output nothing.
- Only include relationships between the tables given below. Do not reference any other tables not in the list.
- Ignore self-referencing relationships (i.e. a table with a column that references itself)

{table_defs}

List the output relationships in child column -> parent column format with no other explanation or output.
Include schema and table names in the column names if possible. For example, if the table was called "main.users" and the column was called "id", the column format must be "main.users.id".
Output:"""


def parse_nlp_table_relationships(output):
    """
    Parse the output of the NLP table relationships prompt.

    **Parameters:**

    * **output** - (*str*) The output of the prompt

    **Returns:**

    (*dict*) - Map child columns to parent columns

    """
    if not output:
        return {}
    child_parent = {}
    for row in output.strip().split("\n"):
        if not row:
            continue
        if "->" not in row:
            warn(f"Invalid row in NLP table relationships output: {row}")
            continue
        child_column, parent_column = [x.strip() for x in row.split("->")]
        child_parent[child_column] = parent_column
    return child_parent


def get_nlp_table_relationships(metadata, table_names):
    """
    Get the NLP table relationships for the given tables. Note: if a table has
    a composite primary key it will be skipped.

    **Parameters:**

    * **metadata** - (*SQLAlchemy Metadata*) The metadata for the database
    * **table_names** - (*list of str*) The names of the tables to get the relationships for

    **Returns:**

    (*dict*) - Map child columns to parent columns

    """
    if not table_names:
        return {}

    table_defs = []

    def get_column_str(c):
        return f"{c.name} ({c.type}) primary_key:{c.primary_key}"

    for table_name in table_names:
        table = metadata.tables[table_name]
        if len(table.primary_key) > 1:
            warn(f"Skipping table {table_name} with composite primary key")
            continue
        table_defs.append(
            f"Table: {table_name}\n"
            "Fields:\n"
            f"{chr(10).join(get_column_str(c) for c in table.columns)}"
        )

    table_defs_str = "\n\n".join(table_defs)

    prompt = PromptTemplate(
        input_variables=["table_defs"],
        template=NLP_TABLE_RELATIONSHIP_PROMPT,
    )
    relationship_chain = build_chain(prompt)

    llm_start = time.time()
    output = relationship_chain.run(table_defs_str).strip(". -\n\r")
    info(output)
    info(f"LLM took {time.time() - llm_start:.3f}s")
    return parse_nlp_table_relationships(output)


NLP_TABLE_COLUMN_PROMPT = """For each column in the following table definition, list the following comma separated:

* The column name
* Whether the column is a metric or dimension
* The aggregation type for the metric (sum or mean), or "NULL" if it is a dimension
* The rounding for "mean" metrics, "NULL" if it is a dimension or "sum" metric

Example output rows:
id,dimension,NULL,NULL
revenue,metric,sum,2
cpc,metric,mean,2

Table definition:
{create_table}
"""


def parse_nlp_table_info(output):
    """Parse the output from the LLM to get the table info

    **Parameters:**

    * **output** - (*str*) The output from the LLM

    **Returns:**

    (*dict*) - A mapping of column names to properties

    """
    res = {}
    for row in output.strip().split("\n"):
        name, type, aggregation, rounding = [x.strip() for x in row.split(",")]
        res[name] = dict(
            type=type,
            aggregation=aggregation if aggregation != "NULL" else None,
            rounding=int(rounding) if rounding != "NULL" else None,
        )
    return res


table_info_chain = None


def get_nlp_table_info(table):
    """Build a langchain chain to get the table info from the LLM

    **Parameters:**

    * **table** - (*SQLAlchemy table*) The table to analyze

    **Returns:**

    (*dict*) - A mapping of column names to properties

    """
    raiseifnot(table.bind, "Table must be bound to an engine")
    create_table = str(CreateTable(table).compile(table.bind)).strip()

    global table_info_chain
    if not table_info_chain:
        prompt = PromptTemplate(
            input_variables=["create_table"], template=NLP_TABLE_COLUMN_PROMPT
        )
        table_info_chain = build_chain(prompt)

    llm_start = time.time()
    output = table_info_chain.run(create_table).strip(". -\n\r")
    info(output)
    info(f"LLM took {time.time() - llm_start:.3f}s")
    return parse_nlp_table_info(output)
